local addonName, addon = ...

local C_Map = C_Map
local C_VignetteInfo = C_VignetteInfo
local CreateFrame = CreateFrame
local CreateFromMixins = CreateFromMixins
local MapCanvasDataProviderMixin = MapCanvasDataProviderMixin
local MapCanvasPinMixin = MapCanvasPinMixin
local WorldMapFrame = WorldMapFrame

local TomCatsVignetteTooltip = TomCatsVignetteTooltip

local maps = { }
local atlasSizes = { }
local vignettes
local interval, minInterval, maxInterval = 1, 1/15, 1
local timeSinceLastUpdate = 0
local vignetteInfoCache = { }

local PIN_STATUS = {
	STATIC = {
		texture = "VignetteKill",
		scaleFactor = 1.00
	},
	SPAWNED = {
		texture = "nazjatar-nagaevent",
		scaleFactor = 1.40
	},
	COMPLETE = {
		texture = "Capacitance-General-WorkOrderCheckmark",
		scaleFactor = 0.90
	}
}

local function ShowHide(element, condition)
	if (condition) then
		element:Show()
		return true
	end
	element:Hide()
	return false
end

local function rescale(pin)
	local atlasName = pin.Texture:GetAtlas()
	if (atlasName) then
		if (not atlasSizes[atlasName]) then
			atlasSizes[atlasName] = { pin.Texture:GetSize() }
		end
		local zoom = pin:GetParent():GetScale()
		local mapData = maps[pin:GetMap()]
		local sizeX = atlasSizes[atlasName][1] / zoom * (pin.scaleFactor or 1) * mapData.iconScale
		local sizeY = atlasSizes[atlasName][2] / zoom * (pin.scaleFactor or 1) * mapData.iconScale
		pin.Texture:SetSize(sizeX, sizeY)
		pin.HighlightTexture:SetSize(sizeX, sizeY)
		pin:SetSize(sizeX, sizeY)
	end
end

local function setPinIcon(pin, status)
	pin.Texture:SetAtlas(status.texture, true)
	pin.HighlightTexture:SetAtlas(status.texture, true)
	pin.scaleFactor = status.scaleFactor
	rescale(pin)
end

local function updateVignettePin(pin)
	local vignette = pin.vignette
	if (vignette) then
		if (pin.isSpawned) then
			setPinIcon(pin, PIN_STATUS.SPAWNED)
		else
			local isCompleted = vignette.isCompleted
			if (isCompleted) then
				setPinIcon(pin, PIN_STATUS.COMPLETE)
			else
				setPinIcon(pin, PIN_STATUS.STATIC)
			end
		end
	end
end

local function OnUpdate(_, elapsed)
	timeSinceLastUpdate = timeSinceLastUpdate + elapsed
	if (timeSinceLastUpdate > interval) then
		timeSinceLastUpdate = 0
		local mapID = C_Map.GetBestMapForUnit("player")
		local affectedMaps
		for mapFrame, mapData in pairs(maps) do
			if (next(mapData.pins) and mapID == mapFrame:GetMapID() and mapFrame:IsShown()) then
				affectedMaps = affectedMaps or { }
				table.insert(affectedMaps, mapData)
			end
		end
		local newInterval
		if (affectedMaps) then
			vignettes = C_VignetteInfo.GetVignettes()
			local vignetteGUIDsByVignetteID
			if (#vignettes ~= 0) then
				newInterval = minInterval
				for _, v in ipairs(vignettes) do
					local vignetteID = vignetteInfoCache[v]
					if (not vignetteID) then
						local vignetteInfo = C_VignetteInfo.GetVignetteInfo(v)
						vignetteID = vignetteInfo.vignetteID
						vignetteInfoCache[v] = vignetteID
					end
					vignetteGUIDsByVignetteID = vignetteGUIDsByVignetteID or { }
					vignetteGUIDsByVignetteID[vignetteID] = vignetteGUIDsByVignetteID[vignetteID] or { }
					table.insert(vignetteGUIDsByVignetteID[vignetteID], v)
				end
			end
			for vignetteID in pairs(affectedMaps[1].pins) do
				local vignetteGUIDs = vignetteGUIDsByVignetteID and vignetteGUIDsByVignetteID[vignetteID]
				for _, mapInfo in ipairs(affectedMaps) do
					local pin = mapInfo.pins[vignetteID]
					local spawned
					if (pin) then
						if (vignetteGUIDs) then
							local coordinates
							if (not coordinates and coordinates ~= false) then
								local idx = C_VignetteInfo.FindBestUniqueVignette(vignetteGUIDs)
								if (idx) then
									coordinates = C_VignetteInfo.GetVignettePosition(vignetteGUIDs[idx], mapID)
								else
									coordinates = false
								end
							end
							if (coordinates) then
								pin.isSpawned = true
								updateVignettePin(pin)
								pin:SetPosition(coordinates:GetXY())
								spawned = true
							end
						end
						if (not spawned) then
							pin.isSpawned = false
							updateVignettePin(pin)
							pin:SetPosition(pin.vignette:GetLocation())
						end
					end
				end
			end
		end
		interval = newInterval or maxInterval
	end
end

TomCatsMapCanvasDataProviderMixin = CreateFromMixins(MapCanvasDataProviderMixin)

function TomCatsMapCanvasDataProviderMixin:OnAdded(owningMap)
	self.owningMap = owningMap
	owningMap:GetPinFrameLevelsManager():
	InsertFrameLevelAbove("PIN_FRAME_LEVEL_TOMCATS", "PIN_FRAME_LEVEL_AREA_POI", 200)
end

function TomCatsMapCanvasDataProviderMixin:RefreshAllData()
	local mapFrame = self:GetMap()
	local mapData = maps[mapFrame]
	if (mapData) then
		local vignetteInfo = addon.getVignettes(mapFrame:GetMapID())
		if (vignetteInfo) then
			for vignetteID, vignette in pairs(vignetteInfo) do
				if (vignette.isVisible) then
					if (not mapData.pins[vignetteID]) then
						ShowHide(mapFrame:AcquirePin("TomCatsMapPinTemplate", vignette), not self.hideAll);
					end
				end
			end
		end
	end
end

function TomCatsMapCanvasDataProviderMixin:OnCanvasScaleChanged()
	local mapData = maps[self:GetMap()]
	for _, pin in pairs(mapData.pins) do
		rescale(pin)
	end
end

function TomCatsMapCanvasDataProviderMixin:OnMapChanged()
	local mapFrame = self:GetMap()
	local mapData = maps[mapFrame]
	for vignetteID, pin in pairs(mapData.pins) do
		mapData.pins[vignetteID] = nil
		pin:Hide()
		mapFrame:RemovePin(pin)
	end
	self:RefreshAllData()
end

local eventFrame = CreateFrame("Frame")

local function setQuestFocus(self, questID)
	local mapFrame = self:GetMap()
	local mapData = maps[mapFrame]
	self.hideAll = questID ~= nil
	for _, pin in pairs(mapData.pins) do
		ShowHide(pin, not self.hideAll)
	end
end

local function setupMapProvider(map, iconScale)
	if (not maps[map]) then
		local provider = CreateFromMixins(TomCatsMapCanvasDataProviderMixin)
		map:AddDataProvider(provider)
		maps[map] = {
			pins = { },
			iconScale = iconScale,
			mapFrame = map
		}
		table.insert(addon.ruleListeners, function()
			provider:OnMapChanged()
		end)
		if (map == WorldMapFrame) then
			map:RegisterCallback("SetFocusedQuestID", setQuestFocus, provider);
			map:RegisterCallback("ClearFocusedQuestID", setQuestFocus, provider);
		end
	end
end

local function OnEvent(_, event, arg1)
	if (event == "ADDON_LOADED") then
		if (addonName == arg1) then
			setupMapProvider(WorldMapFrame, 0.7)
			if (BattlefieldMapFrame) then
				setupMapProvider(BattlefieldMapFrame, 0.7)
			end
		elseif ("Blizzard_BattlefieldMap" == arg1) then
			setupMapProvider(BattlefieldMapFrame, 0.7)
		end
	end
end

eventFrame:SetScript("OnEvent", OnEvent)
eventFrame:SetScript("OnUpdate", OnUpdate)
eventFrame:RegisterEvent("ADDON_LOADED")

TomCatsMapCanvasPinMixin = CreateFromMixins(MapCanvasPinMixin)

function TomCatsMapCanvasPinMixin:OnAcquired(vignette)
	self.vignette = vignette
	local mapData = maps[self:GetMap()]
	mapData.pins[vignette.ID] = self
	self:SetPosition(vignette:GetLocation())
	updateVignettePin(self)
	self:UseFrameLevelType("PIN_FRAME_LEVEL_TOMCATS");
	rescale(self)
end

function TomCatsMapCanvasPinMixin:OnReleased()
	if (self.vignette) then
		maps[self:GetMap()].pins[self.vignette.ID] = nil
	end
	self.vignette = nil
	self.update = nil
	self.isSpawned = nil
	self.scaleFactor = nil
	self.Texture:SetAtlas(nil)
	self.HighlightTexture:SetAtlas(nil)
end

function TomCatsMapCanvasPinMixin:OnMouseEnter()
	TomCatsVignetteTooltip:SetOwner(self)
end

function TomCatsMapCanvasPinMixin:OnMouseLeave()
	TomCatsVignetteTooltip:SetOwner()
end

function TomCatsMapCanvasPinMixin:OnUpdate()
	if (self.update) then self:update() end
end
